
#pragma Ospace

#include <nds.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "_console.h"
#include "_consolewritelog.h"
#include "maindef.h"
#include "memtool.h"
#include "_const.h"

#include "procstate.h"
#include "datetime.h"

#include "glib/glib.h"

#include "fat2.h"
#include "shell.h"

#include "skin.h"
#include "component.h"
#include "sndeff.h"
#include "lang.h"
#include "strpcm.h"
#include "rect.h"

static TProcState *pCurrentProcState;

// -----------------------------

static CglCanvas *pScreenBack;
static s32 ScreenPosY,ShowPosY;

static void ScrollBar_SetPosY(s32 PosY)
{
  ScreenPosY=PosY;
  if(ScreenPosY<0) ScreenPosY=0;
  if(ScreenHeight<ScreenPosY) ScreenPosY=ScreenHeight;
}

// -----------------------------

enum ECompLabels {ECLS_TitleLbl,ECLSCount};
#define CompLabelsCount (ECLSCount)
static TComponentLabel CompLabels[CompLabelsCount];

enum ECompChecks {ECCS_SkipSetupChk,ECCS_BootCheckDiskChk,
                  ECCS_ClickSoundChk,ECCS_EnableFadeEffectChk,ECCS_AutoLastStateChk,
                  ECCS_MoveFolderLockedChk,ECCS_HiddenFilenameExtChk,ECCS_HiddenNotSupportFileTypeChk,
                  ECCS_ShowID3TagChk,ECCS_ScreenSaver_BlackModeChk,ECCS_DisableLRKeyOnPanelClosedChk,
                  ECCS_EnabledFastStartChk,
                  ECCSCount};
#define CompChecksCount (ECCSCount)
static TComponentCheck CompChecks[CompChecksCount];

enum ECompButtons {ECBS_CancelBtn,ECBS_OkBtn,ECBSCount};
#define CompButtonsCount (ECBSCount)
static TComponentButton CompButtons[CompButtonsCount];

static void Setting_Redraw_TransOnly(void)
{
  s32 posy=ShowPosY;
  
  if(posy<ScreenHeight){
    CglB15 *pb15=Custom_GetSkin(ECS_TopMsg);
    pb15->BitBlt(pScreenSub->pCanvas,0,0,ScreenWidth,ScreenHeight-posy,0,0);
  }
  
  pScreenBack->BitBlt(pScreenSub->pCanvas,0,ScreenHeight-posy,ScreenWidth,posy,0,0,false);
  
  pScreenBack->BitBlt(pScreenMain->pBackCanvas,0,0,ScreenWidth,ScreenHeight,0,posy,false);
  
  {
    CglCanvas *pcan=pScreenMain->pBackCanvas;
    const u16 FrameColor=ColorTable.Component.SoftwareScrollBar_Frame;
    const u16 BGColor=ColorTable.Component.SoftwareScrollBar_BG;
    
    s32 w=10,h=ScreenHeight/2;
    s32 x=ScreenWidth-2-w,y=posy/2;
    pcan->SetColor(FrameColor);
    pcan->DrawBox(x,y,w,h);
    
    x+=1; y+=1; w-=2; h-=2;
    u16 *pbuf=pcan->GetScanLine(y);
    pbuf+=x;
    u32 bufsize=pcan->GetWidth()-w;
    for(s32 py=0;py<h;py++){
      for(s32 px=0;px<w;px++){
        u16 col=*pbuf;
        u16 mask=RGB15(30,30,30);
        col=(col&mask)>>1;
        *pbuf++=(BGColor+col)|BIT15;
      }
      pbuf+=bufsize;
    }
  }
  
  ScreenMain_Flip_ProcFadeEffect();
}

static void Setting_Redraw(void)
{
  {
    CglB15 *pb15=Custom_GetSkin(ECS_BG);
    pb15->pCanvas->BitBltFullBeta(pScreenBack);
  }
  
  for(u32 idx=0;idx<CompChecksCount;idx++){
    TComponentCheck *pcc=&CompChecks[idx];
    bool *pchk=(bool*)pcc->UserData;
    if(pchk!=NULL) pcc->Checked=*pchk;
  }
  
  for(u32 idx=0;idx<CompLabelsCount;idx++){
    ComponentLabel_Draw(&CompLabels[idx]);
  }
  for(u32 idx=0;idx<CompChecksCount;idx++){
    ComponentCheck_Draw(&CompChecks[idx]);
  }
  for(u32 idx=0;idx<CompButtonsCount;idx++){
    ComponentButton_Draw(&CompButtons[idx]);
  }
  
  Setting_Redraw_TransOnly();
}

static void CB_CancelBtn_Click(void *pComponentButton)
{
  Sound_Start(WAVFN_Click);
  SetNextProc(ENP_FileList,EPFE_LeftToRight);
}

static void CB_OkBtn_Click(void *pComponentButton)
{
  Sound_Start(WAVFN_Click);
  
  ProcState=*pCurrentProcState;
  ProcState_RequestSave=true;
  ProcState_Save();
  
  SetNextProc(ENP_FileList,EPFE_LeftToRight);
}

static void CB_CheckBox_Click(void *pComponentButton)
{
  Sound_Start(WAVFN_Click);
  TComponentCheck *pcc=(TComponentCheck*)pComponentButton;
  bool *pchk=(bool*)pcc->UserData;
  if(pchk!=NULL) *pchk=!*pchk;
}

static void CompsInit(void)
{
  CglCanvas *pcan=pScreenBack;
  
  for(u32 idx=0;idx<CompLabelsCount;idx++){
    ComponentLabel_Init(&CompLabels[idx],pcan);
    TComponentLabel *pcl=&CompLabels[idx];
  }
  for(u32 idx=0;idx<CompChecksCount;idx++){
    ComponentCheck_Init(&CompChecks[idx],pcan);
    TComponentCheck *pcc=&CompChecks[idx];
    pcc->CallBack_Click=CB_CheckBox_Click;
    pcc->UserData=NULL;
    pcc->pOnIcon=ComponentAlpha_GetSkin(ECSA_CheckOn);
    pcc->pOffIcon=ComponentAlpha_GetSkin(ECSA_CheckOff);
  }
  for(u32 idx=0;idx<CompButtonsCount;idx++){
    ComponentButton_Init(&CompButtons[idx],pcan);
    TComponentButton *pcb=&CompButtons[idx];
  }
  
  // -----------------
  
  u32 TextHeight=glCanvasTextHeight;
  
  u32 x=0,y=0,w=0,h=0;
  
  h=TextHeight;
  
  // -----------------
  
  x=8;
  y+=8;
  
  {
    TComponentLabel *pcl=&CompLabels[ECLS_TitleLbl];
    pcl->pMsgUTF8=Lang_GetUTF8("CS_Title");
    pcl->Rect=CreateRect(x+24,y,ScreenWidth-x,h);
    pcl->isTitle=true;
  }
  
  y+=12*2;
  
  // -----------------
  
  u32 chkw=24,chkh=14;
  
  u32 PaddingY=8;
  
  y+=PaddingY;
  
  TProcState *pcps=pCurrentProcState;
  
  // -----------------
  
  {
    TComponentCheck *pcc=&CompChecks[ECCS_SkipSetupChk];
    pcc->UserData=(u32)&pcps->System.SkipSetup;
    pcc->pMsgUTF8=Lang_GetUTF8("CS_SkipSetup");
    pcc->Rect=CreateRect(x,y,chkw,chkh);
  }
  
  y+=TextHeight*2;
  
  {
    TComponentCheck *pcc=&CompChecks[ECCS_BootCheckDiskChk];
    pcc->UserData=(u32)&pcps->System.BootCheckDisk;
    pcc->pMsgUTF8=Lang_GetUTF8("CS_BootCheckDisk");
    pcc->Rect=CreateRect(x,y,chkw,chkh);
  }
  
  y+=TextHeight*2;
  
  y+=PaddingY;
  
  // -----------------
  
  {
    TComponentCheck *pcc=&CompChecks[ECCS_ClickSoundChk];
    pcc->UserData=(u32)&pcps->System.ClickSound;
    pcc->pMsgUTF8=Lang_GetUTF8("CS_ClickSound");
    pcc->Rect=CreateRect(x,y,chkw,chkh);
  }
  
  y+=TextHeight*2;
  
  {
    TComponentCheck *pcc=&CompChecks[ECCS_EnableFadeEffectChk];
    pcc->UserData=(u32)&pcps->System.EnableFadeEffect;
    pcc->pMsgUTF8=Lang_GetUTF8("CS_EnableFadeEffect");
    pcc->Rect=CreateRect(x,y,chkw,chkh);
  }
  
  y+=TextHeight*2;
  
  {
    TComponentCheck *pcc=&CompChecks[ECCS_AutoLastStateChk];
    pcc->UserData=(u32)&pcps->System.AutoLastState;
    pcc->pMsgUTF8=Lang_GetUTF8("CS_AutoLastState");
    pcc->Rect=CreateRect(x,y,chkw,chkh);
  }
  
  y+=TextHeight*2;
  
  y+=PaddingY;
  
  // -----------------
  
  {
    TComponentCheck *pcc=&CompChecks[ECCS_MoveFolderLockedChk];
    pcc->UserData=(u32)&pcps->FileList.MoveFolderLocked;
    pcc->pMsgUTF8=Lang_GetUTF8("CS_MoveFolderLocked");
    pcc->Rect=CreateRect(x,y,chkw,chkh);
  }
  
  y+=TextHeight*2;
  
  {
    TComponentCheck *pcc=&CompChecks[ECCS_HiddenFilenameExtChk];
    pcc->UserData=(u32)&pcps->FileList.HiddenFilenameExt;
    pcc->pMsgUTF8=Lang_GetUTF8("CS_HiddenFilenameExt");
    pcc->Rect=CreateRect(x,y,chkw,chkh);
  }
  y+=TextHeight*2;
  
  {
    TComponentCheck *pcc=&CompChecks[ECCS_HiddenNotSupportFileTypeChk];
    pcc->UserData=(u32)&pcps->FileList.HiddenNotSupportFileType;
    pcc->pMsgUTF8=Lang_GetUTF8("CS_HiddenNotSupportFileType");
    pcc->Rect=CreateRect(x,y,chkw,chkh);
  }
  
  y+=TextHeight*2;
  
  {
    TComponentCheck *pcc=&CompChecks[ECCS_DisableLRKeyOnPanelClosedChk];
    pcc->UserData=(u32)&pcps->FileList.DisableLRKeyOnPanelClosed;
    pcc->pMsgUTF8=Lang_GetUTF8("CS_DisableLRKeyOnPanelClosed");
    pcc->Rect=CreateRect(x,y,chkw,chkh);
  }
  
  y+=TextHeight*2;
  
  y+=PaddingY;
  
  // -----------------
  
  {
    TComponentCheck *pcc=&CompChecks[ECCS_ShowID3TagChk];
    pcc->UserData=(u32)&pcps->FileList.ShowID3Tag;
    pcc->pMsgUTF8=Lang_GetUTF8("CS_ShowID3Tag");
    pcc->Rect=CreateRect(x,y,chkw,chkh);
  }
  
  y+=TextHeight*2;
  
  {
    TComponentCheck *pcc=&CompChecks[ECCS_ScreenSaver_BlackModeChk];
    pcc->UserData=(u32)&pcps->FileList.ScreenSaver_BlackMode;
    pcc->pMsgUTF8=Lang_GetUTF8("CS_ScreenSaver_BlackMode");
    pcc->Rect=CreateRect(x,y,chkw,chkh);
  }
  
  y+=TextHeight*2;
  
  {
    TComponentCheck *pcc=&CompChecks[ECCS_EnabledFastStartChk];
    pcc->UserData=(u32)&pcps->DPG.EnabledFastStart;
    pcc->pMsgUTF8=Lang_GetUTF8("CS_EnabledFastStart");
    pcc->Rect=CreateRect(x,y,chkw,chkh);
  }
  
  y+=TextHeight*2;
  
  y+=TextHeight;
  
  // -----------------
  
  u32 padx=24;
  
  w=96;
  h=24;
  x=padx;
  y=(ScreenHeight*2)-h-PaddingY;
  
  {
    TComponentButton *pcb=&CompButtons[ECBS_CancelBtn];
    pcb->CallBack_Click=CB_CancelBtn_Click;
    pcb->pIcon=ComponentAlpha_GetSkin(ECSA_Cancel);
    pcb->pMsgUTF8=Lang_GetUTF8("CS_Cancel");
    pcb->Rect=CreateRect(x,y,w,h);
  }
  
  w=72;
  x=ScreenWidth-padx-w;
  
  {
    TComponentButton *pcb=&CompButtons[ECBS_OkBtn];
    pcb->CallBack_Click=CB_OkBtn_Click;
    pcb->pIcon=ComponentAlpha_GetSkin(ECSA_Ok);
    pcb->pMsgUTF8=Lang_GetUTF8("CS_Ok");
    pcb->Rect=CreateRect(x,y,w,h);
  }
}

// -----------------------------

static void CB_KeyDown(u32 Keys)
{
}

static void CB_KeyPress(u32 VsyncCount,u32 Keys)
{
  if((Keys&KEY_B)!=0){
    CB_CancelBtn_Click(NULL);
  }
  
  if((Keys&KEY_A)!=0){
    CB_OkBtn_Click(NULL);
  }
  
  if((Keys&(KEY_UP|KEY_DOWN|KEY_LEFT|KEY_RIGHT))!=0){
    s32 v=0;
    
    if((Keys&(KEY_UP|KEY_LEFT))!=0) v=-1;
    if((Keys&(KEY_DOWN|KEY_RIGHT))!=0) v=1;
    
    ScrollBar_SetPosY(ScreenPosY+(32*v));
  }
}

static void CB_KeyUp(u32 Keys)
{
}

static bool scrmf;
static s32 scrmy;

static bool deskmf;
static TComponentButton *pPressingButton;

static void CB_MouseDown(s32 x,s32 y)
{
  if((ScreenWidth-24)<=x){
    scrmf=true;
    scrmy=y;
    ScrollBar_SetPosY((scrmy*2)-(ScreenHeight/2));
    ShowPosY=ScreenPosY;
    Setting_Redraw_TransOnly();
    return;
  }
  
  deskmf=false;
  
  y+=ScreenPosY;
  
  for(u32 idx=0;idx<CompButtonsCount;idx++){
    TComponentButton *pcb=&CompButtons[idx];
    if(ComponentButton_GetIndexFromPos(pcb,x,y)!=-1){
      pPressingButton=pcb;
      pcb->Pressing=true;
      ComponentButton_Draw(pcb);
      Setting_Redraw();
      deskmf=true;
      return;
    }
  }
  
  deskmf=true;
}

static void CB_MouseMove(s32 x,s32 y)
{
  if(scrmf==true){
    scrmy=y;
    ScrollBar_SetPosY((scrmy*2)-(ScreenHeight/2));
    ShowPosY=ScreenPosY;
    Setting_Redraw_TransOnly();
    return;
  }
  
  if(deskmf==false) return;
  
  y+=ScreenPosY;
  
  for(u32 idx=0;idx<CompButtonsCount;idx++){
    TComponentButton *pcb=&CompButtons[idx];
    if(pcb==pPressingButton){
      if(ComponentButton_GetIndexFromPos(pcb,x,y)==-1){
        if(pcb->Pressing==true){
          pcb->Pressing=false;
          ComponentButton_Draw(pcb);
          Setting_Redraw();
        }
        }else{
        if(pcb->Pressing==false){
          pcb->Pressing=true;
          ComponentButton_Draw(pcb);
          Setting_Redraw();
        }
      }
    }
  }
}

static void CB_MouseUp(s32 x,s32 y)
{
  if(scrmf==true){
    scrmf=false;
    scrmy=y;
    ScrollBar_SetPosY((scrmy*2)-(ScreenHeight/2));
    ShowPosY=ScreenPosY;
    Setting_Redraw_TransOnly();
    return;
  }
  
  if(deskmf==false) return;
  
  deskmf=false;
  
  y+=ScreenPosY;
  
  if(pPressingButton!=NULL){
    pPressingButton->Pressing=false;
    ComponentButton_Draw(pPressingButton);
    
    for(u32 idx=0;idx<CompButtonsCount;idx++){
      TComponentButton *pcb=&CompButtons[idx];
      if(pcb==pPressingButton){
        if(ComponentButton_GetIndexFromPos(pcb,x,y)!=-1){
          ComponentButton_MouseUp(&CompButtons[idx],x,y);
        }
      }
    }
    
    Setting_Redraw();
    pPressingButton=NULL;
    return;
  }
  
  for(u32 idx=0;idx<CompLabelsCount;idx++){
    if(ComponentLabel_MouseUp(&CompLabels[idx],x,y)==true) Setting_Redraw();
  }
  for(u32 idx=0;idx<CompChecksCount;idx++){
    if(ComponentCheck_MouseUp(&CompChecks[idx],x,y)==true) Setting_Redraw();
  }
  
}

static void CB_Start(void)
{
  Sound_Start(WAVFN_Open);
  
  scrmf=false;
  scrmy=0;
  
  deskmf=false;
  pPressingButton=NULL;
  
  pScreenBack=new CglCanvas(NULL,ScreenWidth,ScreenHeight*2,pf15bit);
  ScrollBar_SetPosY(0);
  ShowPosY=ScreenPosY;
  
  pCurrentProcState=(TProcState*)safemalloc(sizeof(TProcState));
  if(pCurrentProcState==NULL){
    _consolePrintf("Fatal error: pCurrentProcState memory overflow.\n");
    ShowLogHalt();
  }
  *pCurrentProcState=ProcState;
  
  CompsInit();
  
  Setting_Redraw();
}

static void CB_VsyncUpdate(u32 VsyncCount)
{
  if(ShowPosY!=ScreenPosY){
    VsyncCount*=4;
    while(VsyncCount!=0){
      VsyncCount--;
      if(ScreenPosY!=ShowPosY){
        if(ScreenPosY<ShowPosY){
          ShowPosY--;
          }else{
          ShowPosY++;
        }
      }
    }
    Setting_Redraw_TransOnly();
  }
}

static void CB_End(void)
{
  if(pCurrentProcState!=NULL){
    safefree(pCurrentProcState); pCurrentProcState=NULL;
  }
  
  if(pScreenBack!=NULL){
    delete pScreenBack; pScreenBack=NULL;
  }
}

void ProcCustom_SetCallBack(void)
{
  CallBack_Start=CB_Start;
  CallBack_VsyncUpdate=CB_VsyncUpdate;
  CallBack_End=CB_End;
  CallBack_KeyDown=CB_KeyDown;
  CallBack_KeyPress=CB_KeyPress;
  CallBack_KeyUp=CB_KeyUp;
  CallBack_MouseDown=CB_MouseDown;
  CallBack_MouseMove=CB_MouseMove;
  CallBack_MouseUp=CB_MouseUp;
}

