
#include <stdio.h>

#include <NDS.h>

#include "_const.h"

#include "plugin.h"
#include "plugin_def.h"

#include "_consoleWriteLog.h"

extern "C" {
#include "jpeglib/jpeglib.h"
}

// CallBack from plugin_dll.cpp

void cbLoadLibrary(void)
{
}

void cbFreeLibrary(void)
{
}

void cbQueryInterfaceLibrary(void)
{
}

// ------------------------------------------------------------------------------------

#include "std.h"

static int FileHandle;

static bool cinfoInit;
ALIGNED_VAR_IN_DTCM struct jpeg_decompress_struct cinfo;

static int imgWidth;
static int imgHeight;
static int imgBits;

static int imgCmpWidth;
static int imgCmpHeight;
static u8 *imgCmpBuf=NULL;

static struct jpeg_error_mgr jerr;

static u32 LastOffsetY;

// ------------------------------------------------------------------------------------

bool JpegStart(int FileHandle)
{
  _consolePrintf("\n");
  _consolePrintf("The Independent JPEG Groups JPEG software\n");
  _consolePrintf("Copyright (C) 1994-1998, Thomas G. Lane.\n");
  _consolePrintf("jpeglib version 6b of 27-Mar-1998.\n");
  _consolePrintf("\n");
  
  MemSet32CPU(0,&cinfo,sizeof(struct jpeg_decompress_struct));
  MemSet32CPU(0,&jerr,sizeof(struct jpeg_error_mgr));
  
  fseek(FileHandle,0,SEEK_SET);
  
  // G[̃nhO
  cinfo.err = jpeg_std_error(&jerr);

  // ȍ~ jpeg CuŃG[ꍇAJďIB

  // \̂̏ݒ
  jpeg_create_decompress(&cinfo);

  // t@C̓nh̐ݒ
  jpeg_stdio_src(&cinfo, (FILE*)FileHandle);

  // t@C̏wb_̓Ǎ
  jpeg_read_header(&cinfo, TRUE);
  
  if(cinfo.progressive_mode==TRUE){
    _consolePrintf("Progressive JPEG is uncorrespondence!!\n");
    jpeg_destroy_decompress(&cinfo);
    cinfoInit=false;
    return(false);
  }

  // 𓀂̊Jn
  jpeg_start_decompress(&cinfo);

  imgWidth=cinfo.image_width;
  imgHeight=cinfo.image_height;
  imgBits=cinfo.output_components*8;
  imgCmpWidth=cinfo.output_width;
  imgCmpHeight=cinfo.output_height;
  
  imgCmpBuf=(u8*)safemalloc(imgCmpWidth*4);
  
  if(imgCmpBuf==NULL){
    _consolePrintf("can not allocate DecompressBuffer. out of memory. (imgCmpWidth=%d)\n",imgCmpWidth);
    jpeg_finish_decompress(&cinfo);
    jpeg_destroy_decompress(&cinfo);
    cinfoInit=false;
    return(false);
  }
  
  cinfoInit=true;
  
  _consolePrint("Jpeg decoder initialized.\n");
  
  return(true);
}

static bool JpegNextLine(u8 *dstbuf)
{
  if(cinfo.output_height<=cinfo.output_scanline){
    _consolePrintf("out of scanline. %d<=%d\n",cinfo.output_height,cinfo.output_scanline);
    return(false);
  }
  
  int res=jpeg_read_scanlines(&cinfo, &imgCmpBuf, 1);
  
  if(res!=1){
    _consolePrintf("error:jpeg_read_scanlines();\n");
    return(false);
  }
  
  u8 *src=imgCmpBuf;
  
  if(imgBits==8){
    for(int x=imgWidth;x!=0;x--){
      u32 b=*src++;
      *dstbuf++=b;
      *dstbuf++=b;
      *dstbuf++=b;
    }
    }else{
    for(int x=imgWidth;x!=0;x--){
      *dstbuf++=*src++;
      *dstbuf++=*src++;
      *dstbuf++=*src++;
    }
  }
  
  return(true);
}

void JpegFree(void)
{
  if(cinfoInit==true){
/*
    int PrgBase=cinfo.output_scanline;
    MWin_ProgressShow("seek lastimage",cinfo.output_height-PrgBase);
    while(cinfo.output_scanline<cinfo.output_height){
      if((cinfo.output_scanline&3)==0) MWin_ProgressSetPos(cinfo.output_scanline-PrgBase);
      if(JpegNextLine(NULL)==false){
        break;
      }
    }
    MWin_ProgressHide();
*/
    
    jpeg_finish_decompress(&cinfo);
    jpeg_destroy_decompress(&cinfo);
    cinfoInit=false;
  }
  
  if(imgCmpBuf!=NULL){
    safefree(imgCmpBuf); imgCmpBuf=NULL;
  }
}

// ------------------------------------------------------------------------------------

bool Start(int _FileHandle)
{
  cinfoInit=false;
  
  if(_FileHandle==0){
    _consolePrintf("FileHandle is NULL\n");
    return(false);
  }
  
  FileHandle=_FileHandle;
  
  if(JpegStart(FileHandle)==false) return(false);
  
  LastOffsetY=-1;
  
  return(true);
}

void Free(void)
{
  FileHandle=0;
  
  JpegFree();
}

void GetBitmap24(u32 LineY,u8 *pBM)
{
  if(imgHeight<=LineY) return;
  
  if((LastOffsetY+1)!=LineY){
    _consolePrintf("Fatal error: Jpeg decode: Not support seek function. (%d,%d)\n",LastOffsetY,LineY);
    ShowLogHalt();
  }
  LastOffsetY=LineY;
  
  if(JpegNextLine(pBM)==false){
    _consolePrintf("Fatal error: Jpeg decode: Decode error.\n");
    ShowLogHalt();
  }
}

s32 GetWidth(void)
{
  return(cinfo.image_width);
}

s32 GetHeight(void)
{
  return(cinfo.image_height);
}

int GetInfoIndexCount(void)
{
  return(2);
}

bool GetInfoStrA(int idx,char *str,int len)
{
  switch(idx){
    case 0: snprintf(str,len,"ImageSize=%dx%dx%dbitColor",imgWidth,imgHeight,imgBits); return(true); break;
    case 1: snprintf(str,len,"ComponentSize=%dx%dpixels",imgCmpWidth,imgCmpHeight); return(true); break;
  }
  return(false);
}

bool GetInfoStrW(int idx,UnicodeChar *str,int len)
{
  return(false);
}

bool GetInfoStrUTF8(int idx,char *str,int len)
{
  return(false);
}

